from .abcattrs import Abstract
from .abcattrs import UndefinedAbstractAttribute
from .abcattrs import abstractattrs
from .abcattrs import check_abstract_class_attributes

__all__ = (
    "Abstract",
    "UndefinedAbstractAttribute",
    "abstractattrs",
    "check_abstract_class_attributes",
)
__version__ = "0.3.2"
